@@ -1,3 +1,24 @@
+/*
+ * This File is part of Davix, The IO library for HTTP based protocols
+ * Copyright (C) CERN 2013
+ * Author: Adrien Devresse <adrien.devresse@cern.ch>
+ *
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ *
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+ *
+*/
+
 #include <cstdio>
 #include <cstring>
 #include <davix.hpp>
@@ -6,6 +27,7 @@
 #include "copy_internal.hpp"
 #include "delegation/delegation.hpp"
 #include <utils/davix_logger_internal.hpp>
+#include <utils/davix_gcloud_utils.hpp>
 
 using namespace Davix;
 
@@ -97,10 +119,24 @@ void DavixCopyInternal::setPerformanceCallback(DavixCopy::PerformanceCallback ca
 void DavixCopyInternal::copy(const Uri &src, const Uri &dst,
         unsigned nstreams, DavixError **error)
 {
-    std::string nextSrc(src.getString()), prevSrc(src.getString());
-    std::string destination(dst.getString());
+    //std::string nextSrc(src.getString()), prevSrc(src.getString());
+    //std::string destination(dst.getString());
+    std::string nextSrc, prevSrc, destination;
     std::string delegationEndpoint;
     DavixError *internalError = NULL;
+    bool suppressFinalHead = false;
+
+    // set source and destination according to copy method
+    if(parameters->getCopyMode() == CopyMode::Push){
+        nextSrc = src.getString();
+        prevSrc = src.getString();
+        destination = dst.getString();
+    }else if(parameters->getCopyMode() == CopyMode::Pull){
+        nextSrc = dst.getString();
+        prevSrc = dst.getString();
+        destination = src.getString();
+    }
+
 
     // nstreams as string
     char nstreamsStr[16];
@@ -110,6 +146,40 @@ void DavixCopyInternal::copy(const Uri &src, const Uri &dst,
     Davix::RequestParams requestParams(parameters);
     requestParams.setTransparentRedirectionSupport(false);
 
+    size_t start_pos;
+
+    // if destination is s3 endpoint, change prefix to http(s) and pre-sign the request as a PUT
+    if(destination.compare(0,2,"s3") == 0){
+        destination.replace(0, 2, "http");
+        time_t expiration_time = time(NULL) +3600;
+        Davix::HeaderVec vec;
+        Uri tmp;
+        if (parameters->getCopyMode() == CopyMode::Pull)
+            tmp = Davix::S3::tokenizeRequest(requestParams, "GET", destination, vec, expiration_time);
+        else
+            tmp = Davix::S3::tokenizeRequest(requestParams, "PUT", destination, vec, expiration_time);
+        destination = tmp.getString();
+        suppressFinalHead = true;
+    }
+
+    // handle gcloud endpoint as a destination
+    if(destination.compare(0,6,"gcloud") == 0){
+        destination.replace(0, 6, "http");
+
+        Davix::HeaderVec vec;
+        Uri tmp;
+        if (parameters->getCopyMode() == CopyMode::Pull)
+            tmp = Davix::gcloud::signURI(requestParams.getGcloudCredentials(), "GET", destination, vec, 3600);
+        else
+            tmp = Davix::gcloud::signURI(requestParams.getGcloudCredentials(), "PUT", destination, vec, 3600);
+        destination = tmp.getString();
+        suppressFinalHead = true;
+    }
+
+    if(destination.compare(0, 3, "dav") == 0) {
+        destination.replace(0, 3, "http");
+    }
+
     // Perform COPY hopping through redirections
     HttpRequest* request = NULL;
     do {
@@ -136,8 +206,19 @@ void DavixCopyInternal::copy(const Uri &src, const Uri &dst,
         }
 
         request->setRequestMethod("COPY");
-        request->addHeaderField("Destination", destination);
+        if(parameters->getCopyMode() == CopyMode::Push){
+            request->addHeaderField("Destination", destination);
+        }else if(parameters->getCopyMode() == CopyMode::Pull){
+            request->addHeaderField("Source", destination);
+        }
         request->addHeaderField("X-Number-Of-Streams", nstreamsStr);
+
+        // for lcgdm-dav, ask for secure redirection in all cases for COPY
+        request->addHeaderField("Secure-Redirection", "1");
+
+        // for lcgdm-dav -> S3, add NoHead flag to suppress final head-to-close request
+        if(suppressFinalHead)
+            request->addHeaderField("Copy-Flags", "NoHead");
         request->setParameters(requestParams);
         request->beginRequest(&internalError);
         if (internalError) {
@@ -171,8 +252,7 @@ void DavixCopyInternal::copy(const Uri &src, const Uri &dst,
             dlg_id.clear();
         }
 
-
-    } while (request->getAnswerHeader("Location", nextSrc));
+    } while (request->getAnswerHeader("Location", nextSrc) && request->getRequestCode() >= 300 && request->getRequestCode() < 400);
 
     if (!*error) {
         int responseStatus = request->getRequestCode();
@@ -193,9 +273,9 @@ void DavixCopyInternal::copy(const Uri &src, const Uri &dst,
                                    "Could not COPY. The source service does not allow it");
         }
         else if (responseStatus == 400) {
+            std::string err_msg(request->getAnswerContentVec().begin(), request->getAnswerContentVec().end());
         	DavixError::setupError(error, COPY_SCOPE, StatusCode::InvalidArgument,
-							   std::string("Could not COPY. The server rejected the request: ") +
-							   request->getAnswerContent());
+                                   fmt::format("Could not COPY. The server rejected the request: {}", err_msg));
         }
         else if (responseStatus >= 300) {
             std::ostringstream msg;
@@ -231,10 +311,13 @@ void DavixCopyInternal::monitorPerformanceMarkers(Davix::HttpRequest *request,
     PerformanceData performance;
     time_t lastPerfCallback = time(NULL);
 
-    while ((line_len = request->readLine(buffer, sizeof(buffer), &daverr)) >= 0
-            && !daverr)
+    while ((line_len = request->readLine(buffer, sizeof(buffer), &daverr)) >= 0 && !daverr)
     {
         buffer[line_len] = '\0';
+
+        if (line_len > 0)
+            DAVIX_SLOG(DAVIX_LOG_VERBOSE, DAVIX_LOG_GRID, "Received: {}", buffer);
+
         // Skip heading whitespaces
         p = buffer;
         while (*p && p < buffer + sizeof(buffer) && isspace(*p))
@@ -273,6 +356,7 @@ void DavixCopyInternal::monitorPerformanceMarkers(Davix::HttpRequest *request,
         }
         else if (strncasecmp("success", p, 7) == 0)
         {
+            request->discardBody(&daverr);
             break;
         }
         else if (strncasecmp("aborted", p, 7) == 0)
@@ -281,21 +365,15 @@ void DavixCopyInternal::monitorPerformanceMarkers(Davix::HttpRequest *request,
                     "Transfer aborted in the remote end");
             break;
         }
-        else if (strncasecmp("failed", p, 6) == 0)
+        else if (strncasecmp("failed", p, 6) == 0 || strncasecmp("failure", p, 7) == 0)
         {
             Davix::DavixError::setupError(error, COPY_SCOPE, StatusCode::RemoteError,
                     std::string("Transfer failed: ") + p);
             break;
         }
-        else
+        else if(line_len> 0)
         {
-            Davix::DavixError::setupError(error, COPY_SCOPE, StatusCode::SystemError,
-                    std::string("Unexpected message from remote host: ") + p);
-            break;
+            DAVIX_SLOG(DAVIX_LOG_WARNING, DAVIX_LOG_GRID, "Unknown performance marker, ignoring: {}", buffer);
         }
     }
 }
-
-
-
-
