@@ -1,6 +1,7 @@
 /*
  * This File is part of Davix, The IO library for HTTP based protocols
- * Copyright (C) 2013  Adrien Devresse <adrien.devresse@cern.ch>, CERN
+ * Copyright (C) CERN 2013
+ * Author: Adrien Devresse <adrien.devresse@cern.ch>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -22,13 +23,15 @@
 #include "neonrequest.hpp"
 
 #include <utils/davix_logger_internal.hpp>
+#include <utils/davix_gcloud_utils.hpp>
 #include <libs/time_utils.h>
 #include <ne_redirect.h>
 #include <ne_request.h>
 #include <davix_context_internal.hpp>
 #include <neon/neonsession.hpp>
 #include <fileops/fileutils.hpp>
-
+#include <fileops/AzureIO.hpp>
+#include <fileops/S3IO.hpp>
 
 
 
@@ -65,6 +68,8 @@ void configureRequestParamsProto(const Uri &uri, RequestParams &params){
             params.setProtocol(RequestProtocol::AwsS3);
         }else if ( proto.compare(0, 3,"dav") ==0){
             params.setProtocol(RequestProtocol::Webdav);
+        }else if ( proto.compare(0, 6, "gcloud") ==0) {
+            params.setProtocol(RequestProtocol::Gcloud);
         }
     }
 
@@ -146,6 +151,7 @@ NEONRequest::NEONRequest(HttpRequest & h, Context& context, const Uri & uri_req)
     _current( new Uri(uri_req)),
     _orig(_current),
     _number_try(0),
+    _redirects(0),
     _total_read_size(0),
     _last_read(-1),
     _vec(),
@@ -164,7 +170,10 @@ NEONRequest::NEONRequest(HttpRequest & h, Context& context, const Uri & uri_req)
     req_started(false),
     req_running(false),
     _last_request_flag(0),
-    _headers_field(){
+    _headers_field(),
+    _accepted_202_retries(0),
+    _early_termination(0),
+    _early_termination_error(NULL) {
 }
 
 
@@ -173,7 +182,10 @@ NEONRequest::~NEONRequest(){
     resetRequest();
 }
 
-void NEONRequest::eradicateSession(){
+void NEONRequest::cancelSessionReuse(){
+   // if session registered
+   // cancel any re-use of the session
+   //
    if(_neon_sess.get() != NULL){
         DAVIX_SLOG(DAVIX_LOG_DEBUG, DAVIX_LOG_HTTP, "Connection problem: eradicate session");
        _neon_sess->disable_session_reuse();
@@ -194,7 +206,16 @@ int NEONRequest::createRequest(DavixError** err){
        resetRequest();
     }
 
-    boost::shared_ptr<Uri> redir_url = ContextExplorer::SessionFactoryFromContext(_c).redirectionResolve(_request_type, *_current);
+    std::shared_ptr<Uri> redir_url;
+    if(this->params.getTransparentRedirectionSupport()) {
+        redir_url = ContextExplorer::SessionFactoryFromContext(_c).redirectionResolve(_request_type, *_current);
+    }
+
+    // performing an operation which could change the PFN? Clear all cache entries for selected URL
+    if(_request_type == "DELETE" || _request_type == "MOVE") {
+        ContextExplorer::SessionFactoryFromContext(_c).redirectionClean(*_current.get());
+    }
+
     if(redir_url.get() != NULL){
         _current.swap(redir_url);
     }
@@ -202,6 +223,22 @@ int NEONRequest::createRequest(DavixError** err){
     if( instanceSession(err) < 0)
         return -1;
 
+
+    // reconfigure protos
+    configureRequestParamsProto(*_current, params);
+
+    // configure S3 params if needed
+    if(params.getProtocol() == RequestProtocol::AwsS3)
+        configureS3params();
+
+    // configure azure params if needed
+    if(params.getProtocol() == RequestProtocol::Azure)
+        configureAzureParams();
+
+    // configure gcloud params if needed
+    if(params.getProtocol() == RequestProtocol::Gcloud)
+        configureGcloudParams();
+
     _req= ne_request_create(_neon_sess->get_ne_sess(), _request_type.c_str(), _current->getPathAndQuery().c_str());
     configureRequest();
 
@@ -234,15 +271,50 @@ bool NEONRequest::checkTimeout(DavixError **err){
     return false;
 }
 
+static dav_ssize_t iocontext_content_provider(HttpBodyProvider provider, void* udata, void* buffer, dav_size_t size) {
+    return provider(udata, (char*) buffer, size);
+}
+
 void NEONRequest::configureRequest(){
-    // reconfigure protos
-    configureRequestParamsProto(*_current, params);
 
-    std::copy(params.getHeaders().begin(), params.getHeaders().end(), std::back_inserter(_headers_field));
+    // add custom user headers, but make sure they're only added once
+    // in case of a redirect
+    if(!req_running) {
+        std::copy(params.getHeaders().begin(), params.getHeaders().end(), std::back_inserter(_headers_field));
+    }
 
-    // configure S3 params if needed
-    if(params.getProtocol() == RequestProtocol::AwsS3)
-        configureS3params();
+    /* Doing a PUT to Azure? Horrific workaround to enforce the odd two-step
+       custom Azure procedure, even when receiving a redirect. */
+    if(_request_type == "PUT" &&
+         _current.get()->queryParamExists("sig") &&
+         _current.get()->queryParamExists("sr") &&
+         _current.get()->queryParamExists("sp") && !_current.get()->fragmentParamExists("azuremechanism")) {
+
+        IOChainContext iocontext(_c, *_current, &params);
+
+        using std::placeholders::_1;
+        using std::placeholders::_2;
+
+        try {
+          AzureIO azureio;
+
+          if(_fd_content > 0) {
+              azureio.writeFromFd(iocontext, _fd_content, _content_len);
+          }
+          else if(_content_provider.callback) {
+              DataProviderFun provider = std::bind(iocontext_content_provider, _content_provider.callback, _content_provider.udata, _1, _2);
+              azureio.writeFromCb(iocontext, provider, _content_len);
+          }
+          else if(_content_ptr && _content_len > 0) {
+              throw DavixException("neon request", StatusCode::InvalidArgument, "unable to use Azure PUT by providing a direct string");
+          }
+        }
+        catch(DavixException &e) {
+            e.toDavixError(&_early_termination_error);
+        }
+
+        _early_termination = true;
+    }
 
     // setup timeout
     if(_expiration_time.isValid() == false
@@ -256,7 +328,8 @@ void NEONRequest::configureRequest(){
         ne_add_request_header(_req, _headers_field[i].first.c_str(),  _headers_field[i].second.c_str());
     }
     // setup flags
-    ne_set_request_flag(_req, NE_REQFLAG_EXPECT100, _req_flag & RequestFlag::SupportContinue100);
+    ne_set_request_flag(_req, NE_REQFLAG_EXPECT100, params.get100ContinueSupport() &&
+                        (_req_flag & RequestFlag::SupportContinue100));
     ne_set_request_flag(_req, NE_REQFLAG_IDEMPOTENT, _req_flag & RequestFlag::IdempotentRequest);
 
     // configure connexion parameters for PUT request
@@ -275,9 +348,27 @@ void NEONRequest::configureRequest(){
 
 
 void NEONRequest::configureS3params(){
-    HeaderVec vec = _headers_field;
-    S3::signRequest(params, _request_type, *_current, vec);
-    vec.swap(_headers_field);
+    // strange workaround to get S3 compatibility on gcloud to work
+    if(params.getAwsRegion().empty()) {
+        HeaderVec vec = _headers_field;
+        S3::signRequest(params, _request_type, *_current, vec);
+        vec.swap(_headers_field);
+    }
+    else {
+        Uri signed_url = S3::signURI(params, _request_type, *_current, _headers_field, NEON_S3_SIGN_DURATION);
+        _current= std::shared_ptr<Uri>(new Uri(signed_url));
+    }
+}
+
+// TODO: make static?
+void NEONRequest::configureAzureParams(){
+    Uri signed_url = Azure::signURI(params.getAzureKey(), _request_type, *_current, NEON_S3_SIGN_DURATION);
+    _current= std::shared_ptr<Uri>(new Uri(signed_url));
+}
+
+void NEONRequest::configureGcloudParams() {
+    Uri signed_url = gcloud::signURI(params.getGcloudCredentials(), _request_type, *_current, _headers_field, NEON_S3_SIGN_DURATION);
+    _current= std::shared_ptr<Uri>(new Uri(signed_url));
 }
 
 int NEONRequest::processRedirection(int neonCode, DavixError **err){
@@ -304,6 +395,15 @@ int NEONRequest::processRedirection(int neonCode, DavixError **err){
     return end_status;
 }
 
+static dav_ssize_t readFunction(int fd, void* buffer, dav_size_t size) {
+    dav_ssize_t ret = ::read(fd, buffer, size);
+    if(ret < 0) {
+        int myerr = errno;
+        DAVIX_SLOG(DAVIX_LOG_DEBUG, DAVIX_LOG_CHAIN, "Error in readFunction when attempting to read from fd {}: Return code {}, errno: {}", fd, ret, myerr);
+    }
+
+    return ret;
+}
 
 int NEONRequest::startRequest(DavixError **err){
     if( createRequest(err) < 0)
@@ -312,8 +412,11 @@ int NEONRequest::startRequest(DavixError **err){
 }
 
 int NEONRequest::negotiateRequest(DavixError** err){
+    std::string ugrs3post;
+    std::string ugrpluginid;
 
-    const int auth_retry_limit = 5;
+    const uint64_t s3SizeLimit = (1024ull * 1024ull * 1024ull * 5ull);
+    const int auth_retry_limit = params.getOperationRetry();
     int code, status, end_status = NE_RETRY;
     _last_read = -1;
     _total_read_size = 0;
@@ -332,7 +435,7 @@ int NEONRequest::negotiateRequest(DavixError** err){
 
     req_started = req_running = true;
 
-    while(end_status == NE_RETRY && _number_try < auth_retry_limit){
+    while(end_status == NE_RETRY && _number_try <= auth_retry_limit) {
         DAVIX_SLOG(DAVIX_LOG_TRACE, DAVIX_LOG_HTTP, "NEON start internal request");
 
         if( (status = ne_begin_request(_req)) != NE_OK && status != NE_REDIRECT){
@@ -343,13 +446,14 @@ int NEONRequest::negotiateRequest(DavixError** err){
                 && (
                     strstr(ne_get_error(_neon_sess->get_ne_sess()), "Could not") != NULL
                     || strstr(ne_get_error(_neon_sess->get_ne_sess()), "Connection reset by peer") != NULL
+                    || strstr(ne_get_error(_neon_sess->get_ne_sess()), "Broken pipe") != NULL
                     )){
                DAVIX_SLOG(DAVIX_LOG_VERBOSE, DAVIX_LOG_HTTP, "Connection problem, retry");
                _number_try++;
                 return startRequest(err);
             }
 
-            if( status == NE_CONNECT &&  requestCleanup() ){
+            if( (status == NE_CONNECT || status == NE_LOOKUP) &&  requestCleanup() ){
                 DAVIX_SLOG(DAVIX_LOG_VERBOSE, DAVIX_LOG_HTTP, "Impossible to connect to {}, retry from origin {}", _current->getString(), _orig->getString());
                 _number_try++;
                 return startRequest(err);
@@ -358,7 +462,7 @@ int NEONRequest::negotiateRequest(DavixError** err){
             req_started= req_running == false;
             createError(status, err);
 
-            eradicateSession();
+            cancelSessionReuse();
             ContextExplorer::SessionFactoryFromContext(_c).redirectionClean(_request_type, *_orig);
             return -1;
         }
@@ -366,15 +470,77 @@ int NEONRequest::negotiateRequest(DavixError** err){
         _last_read =1;
         code = getRequestCode();
         switch(code){
+            case 202:
+                // azure will reply with 202 when deleting files :(
+                // Only activate retries on GET.
+                if(_request_type != "GET") goto default_label;
+
+                _accepted_202_retries++;
+                if(_accepted_202_retries <= params.getAcceptedRetry()) {
+                  std::cerr << "DAVIX: Received 202-Accepted, sleeping for " <<
+                    params.getAcceptedRetryDelay() << " seconds before retrying. (attempt " <<
+                    _accepted_202_retries << " out of " << params.getAcceptedRetry() << ")" << std::endl;
+                  sleep(params.getAcceptedRetryDelay());
+                  endRequest(NULL);
+                  return startRequest(err);
+                }
+                goto default_label;
             case 301:
             case 302:
             case 303:
             case 307:
+
+                if(getAnswerHeader("x-ugrs3posturl", ugrs3post) &&
+                   getAnswerHeader("x-ugrpluginid", ugrpluginid) &&
+                   !ugrs3post.empty() && (_content_len >= s3SizeLimit || _current->fragmentParamExists("forceMultiPart")) ) {
+                    // Ugly workaround for s3 + multi-part upload + dynafed
+                    DAVIX_SLOG(DAVIX_LOG_DEBUG, DAVIX_LOG_HTTP, "Initiating dynafed-assisted multi-part upload to S3, posturl: {}, pluginid: {}", ugrs3post, ugrpluginid);
+                    IOChainContext iocontext(_c, *_current, &params);
+
+                    using std::placeholders::_1;
+                    using std::placeholders::_2;
+
+                    S3IO s3io;
+                    DataProviderFun provider;
+
+                    if(_fd_content > 0) {
+                        provider = std::bind(readFunction, _fd_content, _1, _2);
+                    }
+                    else if(_content_provider.callback) {
+                        provider = std::bind(iocontext_content_provider, _content_provider.callback, _content_provider.udata, _1, _2);
+                    }
+
+                    requestCleanup();
+                    DavixError::clearError(err);
+                    endRequest(NULL);
+
+                    _early_termination = true;
+                    s3io.performUgrS3MultiPart(iocontext, ugrs3post, ugrpluginid, provider, _content_len, &_early_termination_error);
+
+                    if(!err) return 0;
+                    return -1;
+                }
+
                 if( (end_status = processRedirection(status, err)) <0){
                    return -1;
                 }
+
+                if(_early_termination) {
+                    end_status = 0;
+                    break;
+                }
+
+                _number_try--;
+                _redirects++;
+
+                if(_redirects > NEON_REDIRECT_LIMIT) {
+                    httpcodeToDavixError(code, davix_scope_http_request(), "Too many redirects", err);
+                    return -1;
+                }
+
                 break;
             case 401: // authentification requested, do retry
+            case 403:
                 ne_discard_response(_req);
                 end_status = ne_end_request(_req);
                 _last_read = -1;
@@ -391,9 +557,8 @@ int NEONRequest::negotiateRequest(DavixError** err){
                     }
 
                     if(end_status == NE_OK){
-                            DavixError::setupError(err,davix_scope_http_request(),
-                                                   StatusCode::AuthentificationError, "401 Unauthorized");
-                            return -1;
+                        httpcodeToDavixError(code, davix_scope_http_request(), "", err);
+                        return -1;
                     }
 
                     createError(end_status, err);
@@ -401,7 +566,6 @@ int NEONRequest::negotiateRequest(DavixError** err){
                 }
                 DAVIX_SLOG(DAVIX_LOG_DEBUG, DAVIX_LOG_HTTP, "(NEON) {} code -> request again ", code);
                 break;
-            case 403:
             case 501:
                  // cleanup redirection
                 _number_try++;
@@ -410,9 +574,13 @@ int NEONRequest::negotiateRequest(DavixError** err){
                     endRequest(NULL);
                     return startRequest(err);
                 }
-		end_status = 0;
+                end_status = 0;
                 break;
             default:
+            default_label:
+                if(code >= 400) {
+                    httpcodeToDavixError(code, davix_scope_http_request(), "", err);
+                }
                 end_status = 0;
                 break;
 
@@ -420,13 +588,12 @@ int NEONRequest::negotiateRequest(DavixError** err){
         _number_try++;
     }
 
-    if(_number_try >= auth_retry_limit){
-        if(_neon_sess.get() != NULL && _neon_sess->getLastError(err)){
-            DavixError::setupError(err,davix_scope_http_request(), StatusCode::AuthenticationError,
-                               "Authentication Error, reach maximum number of attempts");
-        }
+    if(end_status == NE_RETRY) {
+        DavixError::setupError(err,davix_scope_http_request(), StatusCode::AuthenticationError,
+                               "Authentication error, reached maximum number of attempts");
         return -2;
     }
+
     return 0;
 }
 
@@ -456,7 +623,7 @@ bool NEONRequest::requestCleanup(){
 }
 
 int NEONRequest::redirectRequest(DavixError **err){
-    boost::shared_ptr<Uri> old_uri;
+    std::shared_ptr<Uri> old_uri;
     const ne_uri * new_uri = ne_redirect_location(_neon_sess->get_ne_sess());
     if(!new_uri){
         DavixError::setupError(err, davix_scope_http_request(), StatusCode::UriParsingError, "Impossible to get the new redirected destination");
@@ -468,7 +635,7 @@ int NEONRequest::redirectRequest(DavixError **err){
 
     // setup new path & session target
     old_uri = _current;
-    _current= boost::shared_ptr<Uri>(new Uri(dst_uri));
+    _current= std::shared_ptr<Uri>(new Uri(dst_uri));
     ne_free(dst_uri);
     ContextExplorer::SessionFactoryFromContext(_c).addRedirection(_request_type, *old_uri, _current);
 
@@ -584,6 +751,7 @@ dav_ssize_t NEONRequest::readBlock(char* buffer, dav_size_t max_size, DavixError
     _last_read= read_status= ne_read_response_block(_req, buffer, max_size );
     if(read_status <0){
        DavixError::setupError(err, davix_scope_http_request(), StatusCode::ConnectionProblem, "Invalid Read in request");
+       cancelSessionReuse();
        req_running = false;
        return -1;
     }
@@ -632,7 +800,8 @@ dav_ssize_t NEONRequest::readToFd(int fd, dav_size_t read_size, DavixError** err
            }while(write_len >0);
     }
 
-    return (ret>= 0)?total:ret;
+    if(total > 0) return total;
+    return ret;
 }
 
 
@@ -658,7 +827,7 @@ dav_ssize_t NEONRequest::readLine(char* buffer, dav_size_t max_size, DavixError*
        return (ret >= 0)?(ret + n_bytes):-1;
     }
 
-    if( ( ret = readSegment(buffer, max_size, err)) >= 0){
+    if( ( ret = readSegment(buffer, max_size, true, err)) >= 0){
         // search for crlf
         char* p_endline;
         p_endline = std::find(buffer, buffer+ret, '\n');
@@ -672,23 +841,31 @@ dav_ssize_t NEONRequest::readLine(char* buffer, dav_size_t max_size, DavixError*
 }
 
 
-dav_ssize_t NEONRequest::readSegment(char* p_buff, dav_size_t size_read,  DavixError**err){
+dav_ssize_t NEONRequest::readSegment(char* p_buff, dav_size_t size_read, bool stop_at_line_boundary, DavixError**err){
     DavixError* tmp_err=NULL;
     dav_ssize_t ret, tmp_ret;
     dav_size_t s_read= size_read;
     ret = tmp_ret = 0;
     DAVIX_SLOG(DAVIX_LOG_TRACE, DAVIX_LOG_HTTP, "Davix::Request::readSegment: want to read {} bytes ", size_read);
+    bool early_stop = false;
 
     do{
         tmp_ret= readBlock(p_buff, s_read, &tmp_err);
-        if(tmp_ret > 0){ // tmp_ret bytes readed
+
+        if(tmp_ret > 0 && stop_at_line_boundary && std::find(p_buff, p_buff+tmp_ret, '\n') != p_buff+tmp_ret) {
+            early_stop = true;
+        }
+
+        if(tmp_ret > 0){ // tmp_ret bytes read
             ret += tmp_ret;
         }
+
         if(ret > 0 && ret < (dav_ssize_t) size_read){
             p_buff+= tmp_ret;
             s_read -= tmp_ret;
         }
-    }while( tmp_ret > 0
+
+    }while( !early_stop && tmp_ret > 0
             &&  ret < (dav_ssize_t) size_read);
 
     if(tmp_err){
@@ -707,7 +884,7 @@ int NEONRequest::endRequest(DavixError** err){
         if(_last_read != 0){ // if read content, discard it
             DAVIX_SLOG(DAVIX_LOG_TRACE, DAVIX_LOG_HTTP, "(EndRequest)(Libneon) Operation incomplete, kill the connection");
             ne_abort_request(_req);
-            eradicateSession();
+            cancelSessionReuse();
             _last_read = -1;
 
         }
@@ -717,7 +894,7 @@ int NEONRequest::endRequest(DavixError** err){
             createError(status, err);
             if(tmp_err){
                 DAVIX_SLOG(DAVIX_LOG_DEBUG, DAVIX_LOG_HTTP, "(EndRequest)(Libneon) Suppress broken connection {}  -> {} ", tmp_err->getStatus(), tmp_err->getErrMsg());
-                eradicateSession();
+                cancelSessionReuse();
             }
             DavixError::clearError(&tmp_err);
         }
@@ -733,6 +910,10 @@ void NEONRequest::clearAnswerContent(){
 
 
 int NEONRequest::getRequestCode(){
+    if(_early_termination) {
+        if(!_early_termination_error) return 200;
+        return _early_termination_error->getStatus();
+    }
     return ne_get_status(_req)->code;
 }
 
@@ -748,7 +929,7 @@ std::vector<char> & NEONRequest::getAnswerContentVec(){
 }
 
 dav_ssize_t NEONRequest::getAnswerSizeFromHeaders() const{
-    std::string str_file_size;
+    std::string str_file_size="";
     long size=-1;
     if( getAnswerHeader(ans_header_content_length, str_file_size)){
         StrUtil::trim(str_file_size);
@@ -770,6 +951,25 @@ dav_ssize_t NEONRequest::getAnswerSize() const{
     return _ans_size;
 }
 
+
+time_t NEONRequest::getLastModified() const{
+    time_t t=0;
+    std::string str_lastmodified;
+    if( getAnswerHeader("Last-Modified", str_lastmodified)){
+        StrUtil::trim(str_lastmodified);
+        try{
+            t = S3::s3TimeConverter(str_lastmodified);
+        }catch(...){
+            str_lastmodified.clear();
+        }
+    }
+    if( str_lastmodified.empty()){
+       DAVIX_SLOG(DAVIX_LOG_TRACE, DAVIX_LOG_HTTP, "Bad server answer: {} Invalid, impossible to determine last modified time");
+    }
+    return t;
+}
+
+
 bool NEONRequest::getAnswerHeader(const std::string &header_name, std::string &value) const{
     if(_req){
         const char* answer_content = ne_get_response_header(_req, header_name.c_str());
@@ -826,6 +1026,7 @@ void NEONRequest::setRequestBody(HttpBodyProvider provider, dav_size_t len, void
 
 
 void NEONRequest::freeRequest(){
+    DavixError::clearError(&_early_termination_error);
     if(_req != NULL){
         ne_request_destroy(_req);
         _req=NULL;
@@ -863,6 +1064,7 @@ void NEONRequest::createError(int ne_status, DavixError **err){
 
         default:
             neon_generic_error_mapper(ne_status, code, str);
+            break;
     }
     DavixError::setupError(err, davix_scope_http_request(), code, str);
 }
